package gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.geom.Ellipse2D;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;

import numbercruncher.mathutils.BigFunctions;

import util.Event;
import util.Spawn;
import util.Steal;
import util.Usage;
import util.Worker;

@SuppressWarnings("serial")
public class PlotPanel extends JPanel implements PropertyChangeListener {

	private static final Dimension INITIALSIZE = new Dimension(1280, 800);

	private BigDecimal startTime;
	private BigDecimal endTime;
	ArrayList<Worker> workers;
	private ArrayList<Event> criticalPath;

	private double startHeight;
	private double endHeight;
	private double startWidth;
	private double barLength;
	private double barThickness;
	private double heightIncrement;

	private PanelModel canvasModel;

	private boolean paintSteals;
	private boolean paintLeaps;
	private boolean paintSpawns;
	private boolean paintUsages;
	private boolean paintCriticalPath;

	private JPopupMenu popupMenu;

	public PlotPanel(BigDecimal startTime, BigDecimal endTime, ArrayList<Worker> workers, ArrayList<Event> criticalPath) {
		super();
		this.criticalPath = criticalPath;

		setPreferredSize(INITIALSIZE);

		this.workers = workers;
		this.startTime = startTime;
		this.endTime = endTime;

		paintSteals = true;
		paintLeaps = true;
		paintSpawns = true;
		paintUsages = true;
		paintCriticalPath = true;

		popupMenu = new JPopupMenu();
		add(popupMenu);

		updatePaintParameters();

		addMouseListener(new MouseListenerImplementation());
	}

	private void updatePaintParameters() {
		startHeight = getSize().height * 0.1;
		endHeight = getSize().height * 0.9;
		startWidth = 80;
		barLength = getSize().width - startWidth - 40;

		barThickness = getSize().height * 0.25 / workers.size();

		heightIncrement = (endHeight - startHeight + barThickness) / workers.size();
	}

	@Override
	public void paintComponent(Graphics g) {
		super.paintComponent(g);

		if (canvasModel != null) {
			canvasModel.setValue(getPreferredSize().width);
		}

		setBackground(Color.LIGHT_GRAY);
		Graphics2D gd = (Graphics2D) g;

		updatePaintParameters();

		paintWorkerBars(gd);

		if (paintSteals) {

			for (Worker worker : workers) {

				for (Steal steal : worker.steals) {
					gd.setColor(Color.yellow);
					paintSteal(gd, steal);
				}
			}
		}

		if (paintLeaps) {
			for (Worker worker : workers) {

				for (Steal leap : worker.leaps) {
					gd.setColor(Color.MAGENTA);
					paintSteal(gd, leap);
				}
			}
		}

		if (paintSpawns) {

			paintSpawns(gd);
		}

		paintTimeline(gd);

		if (paintCriticalPath) {
			paintCriticalPath(gd);
		}
	}

	private void paintWorkerBars(Graphics2D gd) {
		for (int i = 0; i < workers.size(); i++) {
			Worker worker = workers.get(i);
			gd.setColor(new Color(0x3050f0));
			int heightPosition = (int) (startHeight + heightIncrement * i);
			gd.drawString("Worker " + i, 10, (int) (heightPosition + ((barThickness + 10) / 2)));
			worker.bar = new Rectangle((int) startWidth, heightPosition, (int) barLength, (int) barThickness);
			gd.fill(worker.bar);

			if (paintUsages) {
				paintUsages(gd, worker, heightPosition);
			}
		}
	}

	private void paintUsages(Graphics2D gd, Worker worker, int heightPosition) {
		for (int j = 0; j < worker.usages.size() - 1; j++) {
			Usage usage = worker.usages.get(j);
			Usage nextUsage = worker.usages.get(j + 1);
			int xPos = (int) ((int) startWidth + (barLength * usage.timestamp.subtract(startTime)
					.divide(endTime.subtract(startTime), 10, BigDecimal.ROUND_HALF_UP).doubleValue()));
			int nextXPos = (int) ((int) startWidth + (barLength * nextUsage.timestamp.subtract(startTime)
					.divide(endTime.subtract(startTime), 10, BigDecimal.ROUND_HALF_UP).doubleValue()));
			gd.setColor(usage.color);
			usage.bar = new Rectangle(xPos, (int) (heightPosition + barThickness + 3), nextXPos - xPos, 5);
			gd.fill(usage.bar);
		}
		if (worker.usages.isEmpty()) {
			return;
		}
		Usage usage = worker.usages.get(worker.usages.size() - 1);
		gd.setColor(usage.color);
		int xPos = (int) ((int) startWidth + (barLength * usage.timestamp.subtract(startTime)
				.divide(endTime.subtract(startTime), 10, BigDecimal.ROUND_HALF_UP).doubleValue()));
		usage.bar = new Rectangle(xPos, (int) (heightPosition + barThickness + 3), (int) (startWidth + barLength - xPos), 5);
		gd.fill(usage.bar);

	}

	private void paintSteal(Graphics2D gd, Steal steal) {
		int x = (int) (startWidth + barLength
				* steal.timeStamp.subtract(startTime).divide(endTime.subtract(startTime), 10, RoundingMode.HALF_UP).doubleValue());
		int yOfThief = (int) (startHeight + heightIncrement * steal.thief + barThickness / 2);
		int yOfVictim = (int) (startHeight + heightIncrement * steal.victim + barThickness / 2);

		int distanceWeight = 5;
		int absoluteIndexDistance = Math.abs(steal.thief - steal.victim);
		gd.drawArc(x - absoluteIndexDistance * distanceWeight, Math.min(yOfThief, yOfVictim), absoluteIndexDistance * distanceWeight * 2,
				Math.abs(yOfThief - yOfVictim), 270, 180);

		final int radius = 5;
		gd.setColor(Color.GREEN);
		gd.fillOval(x - radius, yOfThief - radius, radius * 2, radius * 2);

		gd.setColor(Color.RED);
		gd.fillOval(x - radius, yOfVictim - radius, radius * 2, radius * 2);
	}

	private void paintSpawns(Graphics2D gd) {
		for (Worker worker : workers) {
			for (Spawn spawn : worker.spawns) {
				gd.setColor(Color.white);
				int x = (int) (startWidth + barLength
						* spawn.timeStamp.subtract(startTime).divide(endTime.subtract(startTime), 10, RoundingMode.HALF_UP).doubleValue());
				int yOfSpawner = (int) (startHeight + heightIncrement * spawn.spawner + barThickness / 2);

				// The exact numbers here are pretty arbitrary, it's just to get a little bigger circles when there are few workers
				int radius = (int) Math.max(12 / workers.size(), 4);
				spawn.shape = new Ellipse2D.Double(x - radius, yOfSpawner - radius, radius * 2, radius * 2);
				gd.fill(spawn.shape);

				// Paint a small dot in the spawn to differentiate different tasks
				gd.setColor(worker.getColorForSpawn(spawn));
				gd.fillOval(x - 2, yOfSpawner - 2, 4, 4);
			}
		}
	}

	private void paintCriticalPath(Graphics2D gd) {
		Event firstSpawnEvent = getFirstSpawnEvent();

		BigDecimal timediff = endTime.subtract(startTime);
		gd.setColor(Color.red);
		final int lineWidth = 3;

		for (int i = criticalPath.indexOf(firstSpawnEvent); i < criticalPath.size() - 1; i++) {
			final Event firstEvent = criticalPath.get(i);
			final Event secondEvent = criticalPath.get(i + 1);

			int firstY = (int) (startHeight + heightIncrement * firstEvent.worker);
			int firstX = (int) (startWidth + barLength
					* firstEvent.timeStamp.subtract(firstSpawnEvent.timeStamp).divide(timediff, 10, RoundingMode.HALF_UP).doubleValue());
			int secondX = (int) (startWidth + barLength
					* secondEvent.timeStamp.subtract(firstSpawnEvent.timeStamp).divide(timediff, 10, RoundingMode.HALF_UP).doubleValue());

			gd.fillRect(firstX, firstY, secondX - firstX, lineWidth); // paint a horizontal line
			if (firstEvent.worker != secondEvent.worker) {
				// and then a vertical one if needed
				int secondY = (int) (startHeight + heightIncrement * secondEvent.worker);
				gd.fillRect(secondX, Math.min(firstY, secondY), lineWidth, Math.abs(secondY - firstY) + lineWidth);
			}
		}

	}

	private Event getFirstSpawnEvent() {
		for (int i = 0; i < criticalPath.size(); i++) {
			if (criticalPath.get(i).typeOfEvent == 5) {
				return criticalPath.get(i);
			}
		}
		return null; // should never happen
	}

	private void paintTimeline(Graphics2D gd) {
		gd.setColor(Color.blue);
		BigDecimal totalTime = endTime.subtract(startTime);

		final int yStart = getHeight() - 55;
		final int yStop = (int) (startHeight + (heightIncrement * (workers.size() - 1)) + barThickness);

		final int firstX = (int) startWidth;
		final int lastX = (int) (startWidth + barLength);
		final int textHeight = getHeight() - 40;

		// Paint first tick
		gd.drawLine(firstX, yStart, firstX, yStop);
		gd.drawString("0", firstX, textHeight);

		// Paint last tick
		gd.drawLine(lastX, yStart, lastX, yStop);
		gd.drawString(convertToPrettyString(totalTime), (int) (startWidth + barLength - 120), textHeight);

		BigDecimal timeIncrement = calculateTimeIncrement();
		int currentX = firstX;
		for (BigDecimal currentTime = new BigDecimal(0); currentTime.compareTo(totalTime.subtract(timeIncrement.multiply(new BigDecimal(0.75)))) < 0; currentTime = currentTime
				.add(timeIncrement)) {
			currentX = firstX + calculateXAtCurrentTime(currentTime);
			gd.drawLine(currentX, yStart, currentX, yStop);
			gd.drawString(convertToPrettyString(currentTime), (int) (currentX), textHeight);
		}

	}

	private BigDecimal calculateTimeIncrement() {
		BigDecimal totalTime = endTime.subtract(startTime);
		int zoomRatio = Math.max(1, getWidth() / INITIALSIZE.width);
		BigDecimal approximateTimeBetweenEachTick = totalTime.divide(new BigDecimal(zoomRatio), BigDecimal.ROUND_HALF_UP).divide(new BigDecimal(5),
				10, BigDecimal.ROUND_HALF_UP);
		BigDecimal closestWholeNumber = new BigDecimal(Math.pow(10,
				(int) Math.floor(BigFunctions.log(approximateTimeBetweenEachTick, 10).longValue())));

		BigDecimal[] candidates = { closestWholeNumber, closestWholeNumber.multiply(new BigDecimal(2)),
				closestWholeNumber.multiply(new BigDecimal(3)), closestWholeNumber.multiply(new BigDecimal(4)),
				closestWholeNumber.multiply(new BigDecimal(5)), closestWholeNumber.multiply(new BigDecimal(6)),
				closestWholeNumber.multiply(new BigDecimal(7)), closestWholeNumber.multiply(new BigDecimal(8)),
				closestWholeNumber.multiply(new BigDecimal(10)) };

		int index = Arrays.binarySearch(candidates, approximateTimeBetweenEachTick);

		if (index < 0) {
			index = Math.min(-index + 1, candidates.length - 1);
		}

		BigDecimal timeIncrement = candidates[index];
		return timeIncrement;
	}

	private int calculateXAtCurrentTime(BigDecimal currentTime) {
		double percentageOfTotalX = currentTime.divide(endTime.subtract(startTime), 10, BigDecimal.ROUND_HALF_UP).doubleValue();
		double totalX = startWidth + barLength;
		return (int) (totalX * percentageOfTotalX);
	}

	private void setWidth(int width) {
		setPreferredSize(new Dimension(width, getPreferredSize().height));
		setSize(Math.max(width, getParent().getWidth()), getHeight()); // This makes the panel go through with the resize
	}

	@Override
	public void propertyChange(PropertyChangeEvent evt) {
		if (evt.getPropertyName().equals("value")) {
			setWidth((Integer) evt.getNewValue());
		}
	}

	public void setModel(PanelModel model) {
		model.addPropertyChangeListener(this);
		canvasModel = model;
	}

	public void toggleStealPainting() {
		paintSteals = !paintSteals;
	}

	public void toggleLeapPainting() {
		paintLeaps = !paintLeaps;
	}

	public void toggleSpawnPainting() {
		paintSpawns = !paintSpawns;
	}

	public void toggleUsagePainting() {
		paintUsages = !paintUsages;
	}

	public void toggleCriticalPathPainting() {
		paintCriticalPath = !paintCriticalPath;
	}

	private String convertToPrettyString(BigDecimal timestamp) {
		int[] seconds = makeSecondsArray(timestamp);

		String[] units = { "s", "ms", "us", "ns" };
		String s = "";
		for (int i = 0; i < seconds.length; i++) {
			if (seconds[i] > 0) {
				s += seconds[i] + " " + units[i];
				if (i != seconds.length - 1 && seconds[i + 1] > 0) {
					s += ", ";
				}
			}
		}
		return s;
	}

	private int[] makeSecondsArray(BigDecimal timestamp) {
		BigDecimal aThousand = new BigDecimal("1000");

		// [seconds, milliseconds, microseconds,nanoseconds]
		int[] seconds = new int[4];
		for (int i = 3; i >= 0; i--) {
			BigDecimal[] resultAndRemainder = timestamp.divideAndRemainder(aThousand);
			timestamp = resultAndRemainder[0];
			seconds[i] = resultAndRemainder[1].intValue();
		}
		return seconds;
	}

	private void paintClickInfo(Point mouseClickLocation) {
		for (Worker worker : workers) {
			if (worker.bar.contains(mouseClickLocation) && !worker.spawns.isEmpty()) {
				popupMenu.removeAll();

				Spawn clickedSpawn = getClickedSpawn(worker, mouseClickLocation);

				if (clickedSpawn != null) {

					addLabelToPopupMenu(clickedSpawn.taskName);
					addLabelToPopupMenu("Spawn number " + worker.spawns.indexOf(clickedSpawn));

					popupMenu.show(this, mouseClickLocation.x, mouseClickLocation.y);
					return;
				}
			}
		}

		for (Worker worker : workers) {
			for (Usage usage : worker.usages) {
				if (usage.bar.contains(mouseClickLocation)) {
					popupMenu.removeAll();
					addLabelToPopupMenu("Usage percentage: " + usage.percentage);

					popupMenu.show(this, mouseClickLocation.x, mouseClickLocation.y);
					return;
				}
			}
		}
	}

	private Spawn getClickedSpawn(Worker worker, Point mouseClickLocation) {
		double percentageOfTime = (mouseClickLocation.x - startWidth) / barLength;

		final BigDecimal timeStamp = new BigDecimal(startTime.add(endTime.subtract(startTime).multiply(new BigDecimal(percentageOfTime))).toString());
		final Spawn spawn = new Spawn(-1, timeStamp, null);
		final int clickedSpawnIndex = Collections.binarySearch(worker.spawns, spawn);

		Spawn clickedSpawn = worker.getClickedSpawn(clickedSpawnIndex, mouseClickLocation);
		return clickedSpawn;
	}

	private void addLabelToPopupMenu(final String labelText) {
		final JLabel jLabel = new JLabel(labelText);
		jLabel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
		popupMenu.add(jLabel);
	}

	public final class MouseListenerImplementation implements MouseListener {
		@Override
		public void mouseReleased(MouseEvent e) {
		}

		@Override
		public void mousePressed(MouseEvent e) {
		}

		@Override
		public void mouseExited(MouseEvent e) {
		}

		@Override
		public void mouseEntered(MouseEvent e) {
		}

		@Override
		public void mouseClicked(MouseEvent e) {
			paintClickInfo(new Point(e.getX(), e.getY()));
		}
	}
}